"""Test Logging tracking utilities"""
import collections
from pathlib import Path
import tempfile
import logging
import atexit
from pyarrow import fs

from mldock.platform_helpers.mldock.tracking.logging import ExperimentTracker

class TestExperimentTracker:
    """Test Experiment Tracker functionality"""
    @staticmethod
    def __setup_tracker(experiment_logger, fs_base_path):
        """setup tracker"""
        experiment_tracker = ExperimentTracker(
            experiment_logger=experiment_logger,
            base_path='test_tracker',
            experiment_name='test_experiment',
            fs_base_path=fs_base_path,
            file_system=fs.LocalFileSystem()
        )
        # temporarily unregister write artifacts from atexit callbacks to test it
        atexit.unregister(experiment_tracker.write_artifacts)
        return experiment_tracker

    def test_logs_written_to_fs_base_path_success(self, caplog):
        """
            Test logs are created and copied to fs_base_path on write_artifacts
        """

        tmp_dir = tempfile.TemporaryDirectory()
        test_dir = Path(tmp_dir.name, "tracker_output").as_posix()
        logger = logging.getLogger(tmp_dir.name)
        logger.setLevel(logging.DEBUG)

        experiment_tracker = self.__setup_tracker(logger, test_dir)
        experiment_tracker.log(r"first log")

        experiment_tracker.write_artifacts()

        files = [path.name for path in Path(test_dir).glob('**/*') if path.is_file()]

        # cleanup
        tmp_dir.cleanup()
        
        assert (
            collections.Counter(files) == collections.Counter(['logs.txt', 'manifest.json'])
        ), "Failure: Did not create log.txt and manifest.json"

    def test_logs_contain_info_debug_error_success(self, caplog):
        """
            Test logs are created and copied to fs_base_path on write_artifacts
        """

        logger = logging.getLogger("tracker_output")
        logger.setLevel(logging.DEBUG)

        experiment_tracker = self.__setup_tracker(logger, None)
        experiment_tracker.log(r"first log")
        experiment_tracker.log(r"another debug log", level=logging.DEBUG)
        experiment_tracker.log(r"this is a log")
        experiment_tracker.log(r"an error log", level=logging.ERROR)


        for record in caplog.records:
            assert record.levelname in ["INFO", "DEBUG", "ERROR"]

    def test_logs_contain_no_debug_success(self, caplog):
        """
            Test logs are created and copied to fs_base_path on write_artifacts
        """

        logger = logging.getLogger("tracker_output")
        logger.setLevel(logging.INFO)

        experiment_tracker = self.__setup_tracker(logger, None)
        experiment_tracker.log(r"first log")
        experiment_tracker.log(r"another debug log", level=logging.DEBUG)
        experiment_tracker.log(r"this is a log")
        experiment_tracker.log(r"an error log", level=logging.ERROR)


        for record in caplog.records:
            # these should still be logged
            assert record.levelname in ["INFO", "ERROR"]
        for record in caplog.records:
            # these should not be logged
            assert record.levelname not in ["DEBUG"]

    def test_logs_metrics_successful(self, caplog):
        """
            Test logs are created and copied to fs_base_path on write_artifacts
        """

        logger = logging.getLogger("tracker_output")
        logger.setLevel(logging.DEBUG)

        experiment_tracker = self.__setup_tracker(logger, None)
        experiment_tracker.log(r"first log")
        experiment_tracker.log_param('estimators', 5)
        experiment_tracker.log_metric('mae', 0.5)
        experiment_tracker.log(r"another debug log", level=logging.DEBUG)
        experiment_tracker.log_metric('mse', 0.3)
        experiment_tracker.log(r"this is a log")
        experiment_tracker.log_metrics(
            {
                "accuracy": .9,
                "roc": .9
            }
        )
        experiment_tracker.log(r"an error log", level=logging.ERROR)

        metric_logs = []
        for record in caplog.records:
            if "metric:" in record.message:
                metric_logs.append(record.message)
        
        assert len(metric_logs) == 4, "Failure. Metric logs did not match expected"

    def test_logs_params_successful(self, caplog):
        """
            Test logs are created and copied to fs_base_path on write_artifacts
        """

        logger = logging.getLogger("tracker_output")
        logger.setLevel(logging.DEBUG)

        experiment_tracker = self.__setup_tracker(logger, None)
        experiment_tracker.log(r"first log")
        experiment_tracker.log_metric('mae', 0.5)
        experiment_tracker.log(r"another debug log", level=logging.DEBUG)
        experiment_tracker.log(r"this is a log")
        experiment_tracker.log(r"an error log", level=logging.ERROR)
        experiment_tracker.log_param('estimators', 5)
        experiment_tracker.log_params(
            {
                "factors": 10,
                "lr": 0.01
            }
        )

        param_logs = []
        for record in caplog.records:
            if "param:" in record.message:
                param_logs.append(record.message)
        
        assert len(param_logs) == 3, "Failure. Param logs did not match expected"



"""Training entrypoint script"""
from __future__ import print_function, absolute_import
import sys;sys.path.insert(1, ".")  # Do not remove this
from src.trainer import run_training


if __name__ == '__main__':
    # run training workflow
    run_training()



"""
    Serving Endpoint
    This is the file that implements a flask server to do inferences. It's the file that you will
    modify to implement the scoring for your own algorithm.
"""
import numpy as np
from fastapi import FastAPI, Request, HTTPException
import src.prediction as model_serving
from mldock.platform_helpers.mldock.inference.content_encoders import \
    numpy as content_encoders
from src.container.lifecycle import serving_container

app = FastAPI()

@app.on_event("startup")
def startup_event():
    serving_container.startup()

@app.on_event("shutdown")
def shutdown_event():
    serving_container.cleanup()

# serving workflow utilties
@serving_container.wrap
def handler(json_input, environment, logger, **kwargs):
    """
    Handles a request from to the server app
    :param json_input: [dict], request input
    :return: [dict], prediction
    """
    # TODO input transformer
    del json_input

    # TODO model prediction
    pred = np.array(['pred_1', 'pred_2', 'pred_3'])
    logger.info(pred)
    # TODO Add any output processing
    results = pred

    return results

# Serving Endpoints
@app.get('/ping')
async def ping():
    """Determine if the container is working and healthy"""
    return {'message': 'ping!pong!'}


@app.post('/invocations')
async def transformation(request: Request):
    """Do an inference on a single batch of data. In this sample server, we take data as JSON"""
    if request.headers['content-type'] == 'application/json':
        data = request.body
        print(data)

        results = model_serving.handler(
            np.array(data)
        )
        return content_encoders.array_to_json(results)
    else:
        raise HTTPException(
            status_code=415,
            detail={'message': 'This predictor only supports JSON or CSV data'}
        )



"""
    CONTINAER LIFECYCLE MANAGEMENT

    Provides a central place to instantiate container environment, loggers and container assets management objects.
    Call these instantiated objects in to other scripts using a simple import.
"""
import logging
from mldock.platform_helpers.mldock.configuration.environment.gcp import GCPEnvironment

from src.assets import TrainingContainer, ServingContainer

# Init Environment
environment = GCPEnvironment()

# Set debug level
if environment.environment_variables('MLDOCK_LOGS_LEVEL', None) == 'debug':
    log_emit_level=logging.DEBUG
else:
    log_emit_level=logging.INFO

logging.basicConfig(level=log_emit_level)
logger = logging.getLogger('mldock')

# init Training Container
training_container = TrainingContainer(container_environment=environment, container_logger=logger)

# init Serving Container
serving_container = ServingContainer(container_environment=environment, container_logger=logger)



"""PACKAGES COMMANDS"""
import logging
from pathlib import Path
import click

from mldock.api.packages import build_wheels

click.disable_unicode_literals_warning = True
logger = logging.getLogger('mldock')
MLDOCK_CONFIG_NAME = 'mldock.json'


@click.group()
def packages():
    """
    Commands to manage package dependencies in mldock projects
    """

@click.command(context_settings=dict(
    ignore_unknown_options=True,
))
@click.option(
    '--project_directory',
    '--dir',
    '-d',
    help='mldock container project.',
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None
    )
)
@click.argument('pip_wheel_args', nargs=-1, type=click.UNPROCESSED)
def pack(project_directory, pip_wheel_args):
    """
    Command to create a mldock enabled container template
    """

    try:
        if not Path(project_directory, MLDOCK_CONFIG_NAME).exists():
            raise Exception((
                "Path '{}' was not an mldock project. "
                "Confirm this directory is correct, otherwise "
                "create one.".format(project_directory)
            ))

        mldock_src_path = Path(project_directory, 'src')
        container_project_packages_path = Path(mldock_src_path, 'packages').as_posix()

        build_wheels(dist_dir=container_project_packages_path, pip_wheel_args=list(pip_wheel_args))
    except Exception as exception:
        logger.error(exception)
        raise

packages.add_command(pack)



"""Local API methods"""
import os
import logging
from pathlib import Path
import requests
import docker

from mldock.config_managers.cli import \
    CliConfigureManager
from mldock.platform_helpers import utils

logger=logging.getLogger('mldock')

# log formatting
def pretty_build_logs(line: dict):
    """format docker build logs"""
    if line is None:
        return None

    error = line.get('error', None)
    error_detail = line.get('errorDetail', None)

    if error is not None:
        logger.error('{}\n{}'.format(error, error_detail))

    stream = line.get('stream', '')

    if ('Step' in stream) & (':' in stream):
        logger.info(" ==> {MESSAGE}".format(MESSAGE=stream))
    else:
        logger.debug(" ==> {MESSAGE}".format(MESSAGE=stream))

    return None


class DockerLLCManager:
    """Docker APIClient context manager"""
    client = None
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __enter__(self):
        self.client = docker.APIClient(**self.kwargs)
        return self.client

    def __exit__(self, exc_type, exc_value, exc_tb):
        logger.info(self.client.prune_builds())
        self.client.close()

class DockerManager:
    """Docker Client context manager"""
    client = None

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __enter__(self):
        self.client = docker.from_env(**self.kwargs)
        return self.client

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.client.close()

# API methods

def generate_mldock_mount_volume(
    host_prefix: str,
    container_prefix: str,
    working_dir: str,
    base_container_path: str,
    mode: str = 'rw'
):
    """
        generate host and container paths for docker volume mounts.

        args:
            host_prefix (str): host prefix to use relative to working_dir
            container_prefix (str): container prefix to use relative to base_container_path
            working_dir (str): host working directory to build host volumes from
            base_container_path (str): container base path to build container volumes from
            mode (str): (default='rw') docker volume mount read/write mode
    """
    host_path = Path(working_dir, host_prefix).absolute().as_posix()
    container_path = Path(base_container_path, container_prefix).as_posix()
    return host_path, {'bind': container_path, 'mode': mode}

def make_multiple_mldock_mount_volumes(
    volumes: dict,
    working_dir: str,
    base_container_path: str,
    mode: str = 'rw'
):
    """
        Generate multiple docker volume mounts for a collection of volume mount mappings.

        args:
            volumes (dict): volume mount mappings
            working_dir (str): host working directory to build host volumes from
            base_container_path (str): container base path to build container volumes from
            mode (str): (default='rw') docker volume mount read/write mode
    """
    docker_volumes = {}
    for volume in volumes:

        host_mount, container_mount = generate_mldock_mount_volume(
            host_prefix=volume['host_prefix'],
            container_prefix=volume['container_prefix'],
            working_dir=working_dir,
            base_container_path=base_container_path,
            mode=mode
        )

        docker_volumes.update(
            {host_mount: container_mount}
        )
    return docker_volumes

def train_model(working_dir, docker_tag, image_name, entrypoint, **kwargs):
    """
    Trains ML model(s) locally
    :param working_dir: [str], source root directory
    :param docker_tag: [str], the Docker tag for the image
    :param image_name: [str], The name of the Docker image
    """

    config_manager = CliConfigureManager()

    try:
        with DockerManager() as client:
            base_ml_path = kwargs.get('base_ml_path', '/opt/ml')

            container_volumes = make_multiple_mldock_mount_volumes(
                volumes=[
                    {'host_prefix': 'config', 'container_prefix': 'input/config'},
                    {'host_prefix': 'data', 'container_prefix': 'input/data'},
                    {'host_prefix': 'model', 'container_prefix': 'model'},
                    {'host_prefix': 'output', 'container_prefix': 'output'},
                ],
                working_dir=working_dir,
                base_container_path=base_ml_path,
                mode='rw'
            )

            if config_manager.local.get('auth_type', None) is not None:

                credentials_volume = utils.get_sdk_credentials_volume_mount(
                    config_manager.local.get('auth_type')
                )

                container_volumes.update(credentials_volume)

            container = client.containers.run(
                image="{IMAGE}:{TAG}".format(IMAGE=image_name, TAG=docker_tag),
                entrypoint=entrypoint,
                command=kwargs.get('cmd', 'train'),
                environment=kwargs.get('env', {}),
                remove=True,
                tty=True,
                volumes=container_volumes,
                detach=True,
                stream=True
            )
            logs = container.logs(follow=True).decode('utf-8')

            logger.info(logs)
    except (KeyboardInterrupt, SystemExit) as exception:
        logger.error(exception)
        container.kill()
        raise
    except (
        docker.errors.APIError,
        docker.errors.ContainerError,
        docker.errors.ImageNotFound
    ) as exception:
        logger.error(exception)
        raise
    except Exception as exception:
        logger.error(exception)
        raise

def deploy_model(working_dir, docker_tag, image_name, entrypoint, **kwargs):
    """
    Deploys ML models(s) locally
    :param working_dir: [str], source root directory
    :param docker_tag: [str], the Docker tag for the image
    :param image_name: [str], The name of the Docker image
    """

    config_manager = CliConfigureManager()
    try:
        with DockerManager() as client:
            base_ml_path = kwargs.get('base_ml_path', '/opt/ml')

            container_volumes = make_multiple_mldock_mount_volumes(
                volumes=[
                    {'host_prefix': 'config', 'container_prefix': 'input/config'},
                    {'host_prefix': 'data', 'container_prefix': 'input/data'},
                    {'host_prefix': 'model', 'container_prefix': 'model'},
                    {'host_prefix': 'output', 'container_prefix': 'output'},
                ],
                working_dir=working_dir,
                base_container_path=base_ml_path,
                mode='rw'
            )

            if config_manager.local.get('auth_type', None) is not None:

                credentials_volume = utils.get_sdk_credentials_volume_mount(
                    config_manager.local.get('auth_type')
                )

                container_volumes.update(credentials_volume)

            container = client.containers.run(
                image="{IMAGE}:{TAG}".format(IMAGE=image_name, TAG=docker_tag),
                entrypoint=entrypoint,
                command=kwargs.get('cmd', 'serve'),
                environment=kwargs.get('env', {}),
                ports=kwargs.get('port', {8080: 8080}),
                remove=True,
                tty=True,
                volumes=container_volumes,
                auto_remove=True,
                detach=True,
                stream=True
            )

            if kwargs.get('verbose', False):
                logs = container.logs(follow=True).decode('utf-8')

                logger.info(logs)

    except requests.exceptions.ConnectionError as exception:
        logger.info(
            "\nWorker socket connection timed out. "
            "\n\nThis can happen when connection pool is overloaded. "
            "We suggest setting environment variable: "
            "DOCKER_CLIENT_TIMEOUT=120 to give the client more time."
        )

    except (KeyboardInterrupt, SystemExit) as exception:
        logger.error(exception)
        container.kill()
        raise
    except (
        docker.errors.APIError,
        docker.errors.ContainerError,
        docker.errors.ImageNotFound
    ) as exception:
        logger.error(exception)
        raise
    except Exception as exception:
        logger.error(exception)
        raise

def docker_build(
    image_name: str,
    dockerfile_path: str,
    module_path: str,
    requirements_file_path: str,
    **kwargs
):
    """Runs the build executable using docker python sdk.

    Args:
        script_path (str): relative path to script when run on root
        base_path (str):
        image_name (str):
        dockerfile_path (str):
        module_path (str):
        target_dir_name (str):
        requirements_file_path (str):
        docker_tag (str): the docker tag to build
    """

    try:

        with DockerLLCManager(
            max_pool_size=20,
            base_url=os.environ.get('DOCKER_HOST','unix:///var/run/docker.sock')
        ) as client:

            full_image_name = "{IMAGE}:{TAG}".format(
                IMAGE=image_name,
                TAG=kwargs.get('docker_tag', 'latest')
            )
            logs = client.build(
                tag=full_image_name,
                path='.',
                dockerfile=Path(dockerfile_path).as_posix(),
                buildargs={
                    'module_path': Path(module_path).as_posix(),
                    'target_dir_name': Path(kwargs.get("target_dir_name", "src")).as_posix(),
                    'requirements_file_path': Path(requirements_file_path).as_posix(),
                    'container_platform': kwargs.get('container_platform', None)
                },
                nocache=kwargs.get('no_cache', False),
                rm=True,
                decode=True
            )
            for line in logs:
                yield line

    except (KeyboardInterrupt, SystemExit) as exception:
        logger.error(exception)
        raise
    except (
        docker.errors.APIError,
        docker.errors.ContainerError,
        docker.errors.ImageNotFound
    ) as exception:
        logger.error(exception)
        raise
    except (docker.errors.DockerException) as exception:
        if "NewConnectionError" in str(exception):
            raise Exception((
                "Unable to connect to docker daemon. "
                "Set environment variable: DOCKER_HOST"
            )) from exception
        raise
    except Exception as exception:
        logger.error(exception)
        raise

def search_mldock_containers(**kwargs):
    """Performs a docker ps with filters. Incorporating the label to find only MLDOCK Containers"""
    mldock_filters = {}
    client = docker.from_env()
    ancestor = kwargs.get("ancestor", None)
    if ancestor is not None:
        mldock_filters.update({
            "ancestor": ancestor
        })

    label = kwargs.get("label", ["MLDOCK__IS_MLDOCK_CONTAINER=true"])
    mldock_filters.update(
        {
            "label": label
        }
    )
    containers = client.containers.list(filters=mldock_filters)

    return containers



"""MLDOCK CLI MAIN ENTRYPOINT SCIPT"""
# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import os
import click

from mldock.__version__ import __version__
from mldock.__init__ import MLDOCK_LOGO
from mldock.command.container import container
from mldock.command.configure import configure
from mldock.command.local import local
from mldock.command.registry import registry
from mldock.command.templates import templates
from mldock.command.logs import logs
from mldock.command.packages import packages
from mldock.log import configure_logger

click.disable_unicode_literals_warning = True

CLI_VERSION = 'Version: cli {}'.format(__version__)

@click.group()
@click.version_option(message='{}\n{}'.format(MLDOCK_LOGO, CLI_VERSION))
@click.option("-v", "--verbose", count=True, help="Turn on debug logging")
@click.pass_context
def cli(ctx, verbose):
    """
    A CLI that helps put machine learning in places that empower ml developers
    """
    mldock_package_path = os.path.dirname(os.path.realpath(__file__))
    logger=configure_logger('mldock', verbose=verbose)
    logger.info(MLDOCK_LOGO)
    logger.info(CLI_VERSION)
    ctx.obj = {'verbose': verbose, 'mldock_package_path': mldock_package_path, 'logo': MLDOCK_LOGO}

def add_commands(cli_group: click.group):
    """
        add commands to cli group
        args:
            cli (click.group)
    """
    cli_group.add_command(configure)
    cli_group.add_command(container)
    cli_group.add_command(local)
    cli_group.add_command(registry)
    cli_group.add_command(templates)
    cli_group.add_command(logs)
    cli_group.add_command(packages)

add_commands(cli)

if __name__ == '__main__':
    # pylint: disable=no-value-for-parameter
    cli()



"""Registry API methods"""
import logging
import click
import docker

logger=logging.getLogger('mldock')

def compute_progress(current, max_value=10):
    """
        compute the progress bar update

        args:
            current (int): current count of progress bar
            max_value (int): (default=10) maximum length of progress bar
        returns:
            (int): updated progress bar
    """
    if current > max_value:
        multiple = current // max_value
        current = current - max_value * multiple

    return current

def get_layer_state(
    stream: str,
    layer_id: str,
    state: dict = None,
    **kwargs
):
    """
        Computes a status update to the state of layer during a push.
        Additionally, adds a progress bar and flexible fill character.

        args:
            stream (str): stream message
            layer_id (str): stream layer id
            state (dict) (default=None) current state
        kwargs:
            progress (str): (default = '') starting progress bar string
            fill_char (str): (default='=') fill char to use as progress bar
            increment (str): (default=2) progress bar char increment to use
            max_value (str): (default=40) maximum progress bar length
            status_tag (str): (default='pushing') key to use to look for updates
    """
    progress = kwargs.get('progress', '')
    fill_char = kwargs.get('fill_char', '=')
    increment = kwargs.get('increment', 2)
    max_value = kwargs.get('max_value', 40)
    status_tag = kwargs.get('status_tag', 'pushing')

    try:

        if status_tag.lower() == 'pushing':
            # progress = ''
            if stream.lower() == status_tag.lower():
                current_layer = state[layer_id]
                current_progress = (
                    current_layer['progress']
                        .replace("[",'')
                        .replace("]",'')
                        .replace(">",'')
                )
                current_progress = len(current_progress)
                progress = compute_progress(current_progress, max_value=max_value) + increment
                progress = fill_char * progress
                progress = "[{}>]".format(progress)

        state.update(
            {layer_id: {'message': stream, 'progress': progress}}
        )
    except KeyError:
        pass

    return state

def stateful_log_emitter(line: dict, status_tag: str, states: dict = None):
    """
        stateful log emitter

        args:
            line (dict): current record of logs
            status_tag (str): key to use to update state
            states (dict): current state of logs
    """
    if states is None:
        states = {}

    error = line.get('error', None)
    error_detail = line.get('errorDetail', None)
    if error is not None:
        logger.error('{}\n{}'.format(error, error_detail))

    stream = line.get('status','')
    layer_id = line.get('id','')
    progress = line.get('progress','')

    # perform status update, using layer_id as key
    states = get_layer_state(stream, layer_id, states, progress=progress, status_tag=status_tag)

    # clear terminal to allow for friendlier update
    click.clear()

    # emit current state to stdout
    for key, value in states.items():
        logger.info(
            "{KEY}: {MESSAGE} {PROGRESS}".format(
                KEY=key,
                MESSAGE=value['message'],
                PROGRESS=value['progress']
            )
        )

    return states

def push_image_to_repository(
    image_repository: str,
    auth_config: dict,
    tag='latest'
):
    """
        Push image to repository in registry. Using auth_config this will Authenticate client.
    """
    client = docker.from_env()

    push_response = client.images.push(
        repository=image_repository,
        tag=tag,
        stream=True,
        decode=True,
        auth_config=auth_config
    )

    states = {}
    for line in push_response:
        states = stateful_log_emitter(line=line, states=states, status_tag='Pushing')
    return states

def pull_image_from_repository(
    image_repository: str,
    auth_config: dict,
    tag='latest'
):
    """
        Pull image from repository in registry. Using auth_config this will Authenticate client.
    """
    client = docker.from_env()

    pull_response = client.api.pull(
        repository=image_repository,
        tag=tag,
        stream=True,
        decode=True,
        auth_config=auth_config
    )

    states = {}
    for line in pull_response:
        states = stateful_log_emitter(line=line, states=states, status_tag='Downloading')
    return states



"""
    TRAINER SCRIPT - contains custom code for training your model
"""
import os
import argparse
import pickle

import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC

from src.container.lifecycle import training_container

@training_container.wrap
def run_training(environment, logger):
    """Run training worflow"""
    parser = argparse.ArgumentParser()
    # get parameters
    parser.add_argument(
        '--iris', type=str,
        default=os.path.join(environment.input_data_dir, 'iris'),
        help='Expects a .csv file, representing the lookup.'
    )
    parser.add_argument(
        '--test_size', type=float,
        default=environment.hyperparameters.get('test_size', 0.33),
        help='size of test sample to evaluate on.'
    )
    # Get args
    args, _ = parser.parse_known_args()

    data = pd.read_csv(os.path.join(args.iris, 'train.csv'))
    X, y = data.drop(['target'], axis=1), data['target']
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=args.test_size, random_state=123, shuffle=True)

    transformer = StandardScaler()
    transformer.fit(X_train)
    X_train_std = transformer.transform(X_train)
    X_test_std = transformer.transform(X_test)

    logger.info('Running Training')

    model = SVC(kernel='rbf', random_state=0, gamma=.10, C=1.0)
    model.fit(X_train_std, y_train)

    logger.info('Running Evaluation')
    logger.info(f'train_accuracy={model.score(X_train_std, y_train)};')
    logger.info(f'test_accuracy={model.score(X_test_std, y_test)};')

    logger.info('Save Model artifacts')
    transformer_path = os.path.join(environment.model_dir, 'iris/transformer.pkl')

    with open(transformer_path, 'wb') as file_:
        pickle.dump(transformer, file_)

    model_path = os.path.join(environment.model_dir, 'iris/model.pkl')

    with open(model_path, 'wb') as file_:
        pickle.dump(model, file_)



"""LOGS METRICS COMMANDS"""
import logging
import click
from mldock.api.logs import extract_records_with_pattern

click.disable_unicode_literals_warning = True
logger = logging.getLogger('mldock')

MLDOCK_CONFIG_NAME = 'mldock.json'

@click.group()
def metrics():
    """
    Commands to manage and interact with metric logs.
    """

@click.command()
@click.option(
    '--log-path',
    type=str,
    help='a grok pattern'
)
@click.option(
    '--log-file',
    type=str,
    default='logs.txt',
    help='file name'
)
def show(log_path, log_file):
    """show metrics for all runs as a table"""

    pattern=r"metric: %{GREEDYDATA:name}=%{NUMBER:value};"

    extract_records_with_pattern(pattern, log_path, log_file)

def add_commands(cli_group: click.group):
    """
        add commands to cli group
        args:
            cli (click.group)
    """
    cli_group.add_command(show)

add_commands(metrics)



"""Serving entrypoint script"""
from __future__ import print_function
import sys;sys.path.insert(1, ".")  # Do not remove this
import uvicorn
import multiprocessing
from src.container.lifecycle import environment

cpu_count = multiprocessing.cpu_count()

model_server_timeout = environment.environment_variables.int('MLDOCK_SERVER_TIMEOUT', default=60)
model_server_workers = environment.environment_variables.int('MLDOCK_SERVER_WORKERS', default=cpu_count)

if __name__ == '__main__':

    # start server
    uvicorn.run(
        "src.prediction:app",
        host="0.0.0.0",
        port=8080,
        log_level="debug",
        workers=model_server_workers,
        timeout_keep_alive=model_server_timeout
    )



"""PLATFORM HELPERS MLDOCK UTILITIES"""
import json

def _format_key_as_mldock_env_var(key, prefix=None):
    """
        Formats key as mldock env variable.
        Replace ' ' and '-', append mldock prefix
        and lastly transforms to uppercase
    """
    if prefix is not None:
        if not key.lower().startswith(prefix.lower()):
            key = "{PREFIX}_{KEY}".format(PREFIX=prefix, KEY=key)

    key = key.replace(" ", "_").replace("-", "_")
    key = key.upper()
    return key

def covert_if_not_ascii(value):
    """converts a value if it is not a ascii supported str"""
    try:
        value.encode('ascii')
        return value
    except (AttributeError, UnicodeEncodeError):
        return str(value)

def format_dict_for_subprocess(obj: dict):
    """formats dictionary object keys & values to asci"""
    return {
        covert_if_not_ascii(key): covert_if_not_ascii(value)
        for key, value in obj.items()
    }

def _format_dictionary_as_env_vars(obj: dict, group=None):
    """
        format key and values as appropriate environment variables

        namely:
            - upper case
            - replace spaces with "_"
            - replace dashes with "_"
            - add owner prefix
    """
    new_keys = {}
    for _key,_value in obj.items():

        _key = _format_key_as_mldock_env_var(_key, prefix=group)

        if isinstance(_value, dict):
            _value = json.dumps(_value)

        new_keys.update(
            {
                _key: _value
            }
        )
    return new_keys

def collect_mldock_environment_variables(**env_vars):
    """Collect and format mldock environment variables"""
    return _format_dictionary_as_env_vars(
        env_vars,
        group='mldock'
    )

def _make_sagemaker_input_data_configs(data_channels: list):
    """restructures a list of data names as sagemaker input data configs"""
    return {name: {} for name in data_channels}

def _extract_data_channels_from_mldock(mldock_data: list):

    data_channels = [
        data_config['channel'] for data_config in mldock_data
    ]

    return _make_sagemaker_input_data_configs(data_channels=data_channels)



"""LOCAL COMMANDS"""
import os
import sys
from pathlib import Path
import logging
import subprocess
import click

from mldock.config_managers.cli import \
    CliConfigureManager
import mldock.api.predict as predict_request
from mldock.terminal import \
    ChoiceWithNumbers, style_dropdown, style_2_level_detail, \
        ProgressLogger, pretty_build_logs
from mldock.platform_helpers.mldock import utils as mldock_utils
from mldock.config_managers.container import MLDockConfigManager
from mldock.api.local import \
    search_mldock_containers, docker_build, train_model, deploy_model

click.disable_unicode_literals_warning = True
logger = logging.getLogger('mldock')
MLDOCK_CONFIG_NAME = 'mldock.json'

def run_script_interactively(cmd, **kwargs):
    """Basic runner script for local interactive script execution"""
    output = subprocess.check_output(
        cmd,
        **kwargs
    )
    logger.info(output)

def python_executable():
    """Return the real path for the Python executable, if it exists.
    Return RuntimeError otherwise.
    Returns:
        (str): The real path of the current Python executable.
    """
    if not sys.executable:
        raise RuntimeError("Failed to retrieve the real path for the Python executable binary")
    return sys.executable

@click.group()
def local():
    """
    Commands for local development
    """

@click.command()
@click.option(
    '--project_directory',
    '--dir',
    '-d',
    help='mldock container project.',
    required=True,
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None
    )
)
@click.option('--no-cache', help='builds container from scratch', is_flag=True)
@click.option('--tag', help='docker tag', type=str, default='latest')
@click.option('--stage', help='environment to stage.')
def build(project_directory, no_cache, tag, stage):
    """Command to build container locally
    """
    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(project_directory, MLDOCK_CONFIG_NAME)
    )
    # get mldock_module_dir name
    mldock_config = mldock_manager.get_config()
    image_name = mldock_config.get("image_name", None)
    template_name = mldock_config.get("template", None)
    container_dir = mldock_config.get("container_dir", None)
    module_path = os.path.join(
        project_directory,
        mldock_config.get("mldock_module_dir", "src"),
    )
    dockerfile_path = os.path.join(
        project_directory,
        mldock_config.get("mldock_module_dir", "src"),
        container_dir,
        "Dockerfile"
    )
    requirements_file_path = os.path.join(
        project_directory,
        mldock_config.get("requirements_dir", "requirements.txt")
    )

    with ProgressLogger(
        group='Stages',
        text='Retrieving Stages',
        spinner='dots',
        on_success='Stages Retrieved'
    ) as spinner:
        stages = mldock_config.get("stages", None)

        if stage is not None:
            tag = stages[stage]['tag']

        if tag is None:
            raise Exception("tag is not valid. Either choose a stage or set a tag manually")

    try:
        with ProgressLogger(group='Build', text='Building', spinner='dots') as spinner:
            logs = docker_build(
                image_name=image_name,
                dockerfile_path=dockerfile_path,
                module_path=module_path,
                target_dir_name=mldock_config.get("mldock_module_dir", "src"),
                requirements_file_path=requirements_file_path,
                no_cache=no_cache,
                docker_tag=tag,
                container_platform=template_name
            )
            for line in logs:

                spinner.info(pretty_build_logs(line=line))

                spinner.start()

    except Exception as exception:
        logger.error(exception)
        raise

@click.command()
@click.option(
    '--payload',
    help='Path to payload',
    required=True,
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None
    )
)
@click.option(
    '--content-type',
    default='application/json',
    help='format of payload',
    type=click.Choice(
        ['application/json', 'text/csv', 'image/jpeg'],
        case_sensitive=False
    )
)
@click.option(
    '--host',
    help='host url at which model is served',
    type=str,
    default='http://127.0.0.1:8080/invocations'
)
def predict(payload, content_type, host):
    """
    Command to execute prediction request against ml endpoint
    """
    with ProgressLogger(group='Predict', text='Running Request', spinner='dots'):
        if payload is None:
            logger.info("\nPayload cannot be None. Please provide path to payload file.")
        else:
            if content_type in ['application/json']:
                pretty_output = predict_request.send_json(filepath=payload, host=host)
            elif content_type in ['text/csv']:
                pretty_output = predict_request.send_csv(filepath=payload, host=host)
            elif content_type in ['image/jpeg']:
                pretty_output = predict_request.send_image_jpeg(filepath=payload, host=host)
            else:
                raise Exception("Content-type is not supported.")
            logger.info(pretty_output)

@click.command()
@click.option(
    '--project_directory',
    '--dir',
    '-d',
    help='mldock container project.',
    required=True,
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None
    )
)
@click.option(
    '--params',
    '-p',
    help='(Optional) Hyperparameter override when running container.',
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True
)
@click.option(
    '--env_vars',
    '-e',
    help='(Optional) Environment Variables override when running container.',
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True
)
@click.option('--tag', help='docker tag', type=str, default='latest')
@click.option('--stage', help='environment to stage.')
@click.option('--interactive', help='run workflow without docker', is_flag=True)
def train(project_directory, **kwargs):
    """
    Command to run training locally on localhost
    """
    params = kwargs.get('params', None)
    env_vars = kwargs.get('env_vars', None)
    tag = kwargs.get('tag', None)
    stage = kwargs.get('stage', None)
    interactive = kwargs.get('interactive', False)

    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(project_directory, MLDOCK_CONFIG_NAME)
    )
    # get mldock_module_path name
    mldock_config = mldock_manager.get_config()
    image_name = mldock_config.get("image_name", None)

    with ProgressLogger(
        group='Stages',
        text='Retrieving Stages',
        spinner='dots',
        on_success='Stages Retrieved'
    ) as spinner:
        stages = mldock_config.get("stages", 'null')

        if stage is not None:
            tag = stages[stage]['tag']

        if tag is None:
            raise Exception("tag is not valid. Either choose a stage or set a tag manually")

    with ProgressLogger(
        group='Environment',
        text='Retrieving Environment Varaiables',
        spinner='dots',
        on_success='Environment Ready'
    ) as spinner:
        project_env_vars = mldock_config.get('environment', {})
        for env_var in env_vars:
            key_, value_ = env_var
            project_env_vars.update(
                {key_: value_}
            )

        hyperparameters = mldock_config.get('hyperparameters', {})

        # override hyperparameters
        for param in params:
            key_, value_ = param
            hyperparameters.update(
                {key_: value_}
            )

        env_vars = mldock_utils.collect_mldock_environment_variables(
            stage=stage,
            hyperparameters=hyperparameters,
            **project_env_vars
        )

        config_manager = CliConfigureManager()
        env_vars.update(
            config_manager.local.get('environment', {})
        )

    with ProgressLogger(
        group='Training',
        text='Running Training',
        spinner='dots'
    ) as spinner:

        spinner.info("Training Environment = {}".format(env_vars))
        spinner.start()
        if interactive:
            spinner.info("Running interactively")
            spinner.start()
            # just run python src/container/prediction/serve.py
            base_ml_path = project_directory
            # must update /opt/ml working directory before running
            # perhaps setting from environment would be the best
            env_vars.update({
                'MLDOCK_BASE_DIR': Path(base_ml_path).absolute().as_posix(),
                'MLDOCK_INPUT_DIR': '.',
            })

            # subprocess only supports 'ascii' supported str for environment variables
            env_vars = mldock_utils.format_dict_for_subprocess(env_vars)
            script_path = 'src/container/training/train.py'
            run_script_interactively(
                [python_executable(), script_path],
                cwd=base_ml_path,
                env=env_vars
            )
        else:
            spinner.info("Running docker container")
            spinner.start()
            train_model(
                working_dir=project_directory,
                docker_tag=tag,
                image_name=image_name,
                entrypoint="src/container/executor.sh",
                cmd="train",
                env=env_vars
            )


@click.command()
@click.option(
    '--project_directory',
    '--dir',
    '-d',
    help='mldock container project.',
    required=True,
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None
    )
)
@click.option(
    '--params',
    '-p',
    help='(Optional) Hyperparameter override when running container.',
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True
)
@click.option(
    '--env_vars',
    '-e',
    help='(Optional) Environment Variables override when running container.',
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True
)
@click.option('--tag', help='docker tag', type=str, default='latest')
@click.option('--port', help='host url at which model is served', type=str, default='8080')
@click.option('--stage', help='environment to stage.')
@click.option('--interactive', help='run workflow without docker', is_flag=True)
@click.pass_obj
def deploy(obj, project_directory, **kwargs):
    """
    Command to deploy ml container on localhost
    """
    params = kwargs.get('params', None)
    env_vars = kwargs.get('env_vars', None)
    tag = kwargs.get('tag', None)
    port = kwargs.get('port', None)
    stage = kwargs.get('stage', None)
    interactive = kwargs.get('interactive', False)

    verbose = obj.get('verbose', False)
    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(project_directory, MLDOCK_CONFIG_NAME)
    )
    # get mldock_module_path name
    mldock_config = mldock_manager.get_config()
    image_name = mldock_config.get("image_name", None)

    with ProgressLogger(
        group='Stages',
        text='Retrieving Stages',
        spinner='dots',
        on_success='Stages Retrieved'
    ) as spinner:
        stages = mldock_config.get("stages", None)

        if stage is not None:
            tag = stages[stage]['tag']

        if tag is None:
            raise Exception("tag is not valid. Either choose a stage or set a tag manually")

    project_env_vars = mldock_config.get('environment', {})

    for env_var in env_vars:
        key_, value_ = env_var
        print(key_, value_)
        project_env_vars.update(
            {key_: value_}
        )

    hyperparameters = mldock_config.get('hyperparameters', {})

    # override hyperparameters
    for param in params:
        key_, value_ = param
        print(key_, value_)
        hyperparameters.update(
            {key_: value_}
        )

    env_vars = mldock_utils.collect_mldock_environment_variables(
        stage=stage,
        hyperparameters=mldock_config.get('hyperparameters', {}),
        **project_env_vars
    )

    config_manager = CliConfigureManager()
    env_vars.update(
        config_manager.local.get('environment', {})
    )

    with ProgressLogger(
        group='Deploy',
        text='Deploying model to {} @ localhost'.format(port),
        spinner='dots'
    ) as spinner:
        spinner.info("Deployment Environment = {}".format(env_vars))
        spinner.start()
        if interactive:
            spinner.info("Running interactively")
            spinner.start()
            # just run python src/container/prediction/serve.py
            base_ml_path = project_directory
            # must update /opt/ml working directory before running
            # perhaps setting from environment would be the best
            # check out the python runner from sagemaker_training
            env_vars.update({
                'MLDOCK_BASE_DIR': Path(base_ml_path).absolute().as_posix(),
                'MLDOCK_INPUT_DIR': '.'
            })

            # subprocess only supports 'ascii' supported str for environment variables
            env_vars = mldock_utils.format_dict_for_subprocess(env_vars)

            script_path = 'src/container/prediction/serve.py'

            run_script_interactively(
                [python_executable(), script_path],
                cwd=base_ml_path,
                env=env_vars
            )
        else:
            spinner.info("Running docker container")
            spinner.start()

            deploy_model(
                working_dir=project_directory,
                docker_tag=tag,
                image_name=image_name,
                port={8080: port},
                entrypoint="src/container/executor.sh",
                cmd="serve",
                env=env_vars,
                verbose=verbose
            )

@click.command()
def stop():
    """
    Command to stop ml container on localhost
    """

    try:
        containers = search_mldock_containers()

        tags = [
            style_2_level_detail(
                major_detail=c.image.tags[0],
                minor_detail=c.name
             ) for c in containers
        ] + ['abort']
        # newline break
        click.echo("")
        click.secho("Running MLDock containers:", bg='blue', nl=True)
        container_tag = click.prompt(
            text=style_dropdown(group_name="container name", options=tags, default='abort'),
            type=ChoiceWithNumbers(tags, case_sensitive=False),
            show_default=False,
            default='abort',
            show_choices=False
        )

        for container in containers:
            for image_tag in container.image.tags:
                if image_tag in container_tag:
                    # make newline
                    click.echo("")
                    project_container_tag = click.style(container_tag, fg='bright_blue')
                    if click.confirm(
                        "Stop the running container = {}?".format(
                            project_container_tag)
                    ):
                        container.kill()

                    break

    except Exception as exception:
        logger.error(exception)
        raise


def add_commands(cli_group: click.group):
    """
        add commands to cli group
        args:
            cli (click.group)
    """
    cli_group.add_command(build)
    cli_group.add_command(predict)
    cli_group.add_command(train)
    cli_group.add_command(deploy)
    cli_group.add_command(stop)

add_commands(local)



"""LOGS ERRORS COMMANDS"""
import logging
from pathlib import Path
import click
from clickclick import choice

from mldock.api.logs import parse_grok, get_all_file_objects, infer_filesystem_type

click.disable_unicode_literals_warning = True
logger = logging.getLogger('mldock')

def reset_terminal():
    """clears the terminal view frame"""
    click.clear()

@click.group()
def errors():
    """
    Commands to manage and interact with errors logs.
    """

@click.command()
@click.option(
    '--log-path',
    type=str,
    help='a grok pattern'
)
@click.option(
    '--log-file',
    type=str,
    default='logs.txt',
    help='file name'
)
@click.pass_obj
def show(obj, log_path, log_file):
    """show errors for all runs as a table"""

    pattern = r'Exception during %{WORD:script}\: %{GREEDYDATA:msg}'
    file_system, log_path = infer_filesystem_type(log_path)

    logs = get_all_file_objects(log_path, log_file, file_system)

    log_runs = [Path(log).parents[0].name for log in logs]

    state = choice('Select a run', log_runs, default=None)

    for log in logs:
        if Path(log).parents[0].name == state:
            log_file_path = log
            break

    metadata = parse_grok(log_file_path, pattern, file_system)

    reset_terminal()
    logger.info(obj['logo'])
    print(f"Script => {metadata['script']}")
    print(f"Exception:\n\n{metadata['msg']}")


errors.add_command(show)



"""Predict Request API utilties"""
import io
import json
import requests

def execute_request(url, headers, data):
    """compiles and executes request to API"""
    return requests.post(
        url=url,
        headers=headers,
        data=data
    )

def send_json(filepath, host):
    """Send a json payload in request providing a filepath."""
    with open(filepath, 'r') as file_:
        data = json.load(file_)

    response = execute_request(
        url=host,
        headers={'Content-Type': 'application/json'},
        data=json.dumps(data)
    )
    if response.status_code != 200:
        raise requests.exceptions.RequestException(
            "error ({}): {}".format(
                response.status_code,
                response.raise_for_status()
            )
        )

    return "{}".format(
        json.dumps(
            response.json(),
            indent=4,
            separators=(',', ': '),
            sort_keys=True
        )
    )

def send_csv(filepath, host):
    """Send a json payload in request providing a filepath."""
    with open(filepath, 'r') as file_:
        data = file_.read()

    response = execute_request(
        url=host,
        headers={'Content-Type': 'text/csv'},
        data=data
    )

    if response.status_code != 200:
        raise requests.exceptions.RequestException(
            "error ({}): {}".format(
                response.status_code,
                response.raise_for_status()
            )
        )
    return response.text

def send_image_jpeg(filepath, host):
    """Send a image jpeg payload as bytes in request providing a filepath."""

    with open(filepath, 'rb') as file_:
        byte_content = io.BytesIO(file_.read())

    response = execute_request(
        url=host,
        headers={'Content-Type': 'image/jpeg'},
        data=byte_content
    )

    if response.status_code != 200:
        raise requests.exceptions.RequestException(
            "error ({}): {}".format(
                response.status_code,
                response.raise_for_status()
            )
        )
    return "{}".format(
        json.dumps(
            response.json(),
            indent=4,
            separators=(',', ': '),
            sort_keys=True
        )
    )



"""
    Config managers for Configure commands

    Config managers used in command/configure
"""
import sys
import json
import logging
from pathlib import Path
import yaml
import click
from mldock.terminal import ChoiceWithNumbers, style_dropdown
from mldock.config_managers.core import BaseConfigManager
from mldock.platform_helpers import utils

logger = logging.getLogger('mldock')
MLDOCK_CLI_CONFIG="./.mldock/config"

class CliConfigureManager(BaseConfigManager):
    """CLI Configure Manager for mldock
    """

    config = {}

    def __init__(
        self,
        filepath: str = MLDOCK_CLI_CONFIG,
        create: bool = False
    ):
        super().__init__()
        self.filepath = filepath

        self.config = self.load_config(self.filepath, create=create)

    def reset(self, config_type):
        """reset given configuration by config key"""
        click.secho("Dropping {} configuration".format(config_type), bg='blue')
        self.config.pop(config_type, None)

    def check_if_exists_else_create(self, file_name: str, create: bool):
        """check that mldock fileexists"""
        if not self.file_exists(file_name):
            if create:
                # deal with possiblity of nested directory
                utils._mkdir(Path(file_name).parents[0])
                # create file
                self.touch(file_name)
            else:
                logger.error((
                    "No MLDOCK CLI config found at '{CONFIG_PATH}/'. "
                    "To create run: 'mldock configure init'".format(
                        CONFIG_PATH=Path(file_name).parents[0]
                    )
                ))
                sys.exit(1)

    def setup_local_config(self):
        """setup and prompt user for local configuration"""
        click.secho("Setup local CLI configuration", bg='blue')
        if self.config.get('local') is None:
            self.config['local'] = {}
        self.ask_for_container_auth_type()
        self.ask_for_environment(config_type='local')

    def setup_templates_config(self):
        """setup and prompt user for templates configuration"""
        click.secho("Setup templates CLI configuration", bg='blue')
        if self.config.get('templates') is None:
            self.config['templates'] = {}
        self.ask_for_template_server()
        self.ask_for_template_root()

    def setup_workspace_config(self):
        """setup and prompt user for workspace configuration"""
        click.secho("Setup workspace CLI configuration", bg='blue')
        if self.config.get('workspace') is None:
            self.config['workspace'] = {}
        self.ask_for_environment(config_type='workspace')

    def ask_for_container_auth_type(self):
        """prompt user for platform name
        """

        click.secho("Set local authentication type", bg='blue', nl=True)
        options = ['gcloud', 'awscli']

        auth_type = click.prompt(
            text=style_dropdown(
                group_name="authentication type",
                options=options,
                default=self.config.get('auth_type', None)
            ),
            type=ChoiceWithNumbers(options, case_sensitive=False),
            show_default=True,
            default=self.config.get('auth_type', None),
            show_choices=False
        )

        self.config['local'].update({
            'auth_type': auth_type
        })

    def ask_for_environment(self, config_type):
        """
            prompt user for environment variables

            args:
                config_type (str): config key that points to configure config type
        """
        config = self.config[config_type].get('environment', {})
        environment_config_manager = EnvironmentConfigManager(config=config)
        environment_config_manager.ask_for_env_vars()
        self.config[config_type].update({
            'environment': environment_config_manager.get_config()
        })

    def ask_for_template_server(self):
        """prompt user to set template server tracking
        """

        click.secho("Set template server type", bg='blue', nl=True)
        options = ['local', 'github']

        server_type = click.prompt(
            text=style_dropdown(
                group_name="template server type",
                options=options,
                default=self.config.get('server_type', None)
            ),
            type=ChoiceWithNumbers(options, case_sensitive=False),
            show_default=True,
            default=self.config.get('server_type', None),
            show_choices=False
        )

        self.config['templates'].update({
            'server_type': server_type
        })

    def ask_for_template_root(self):
        """prompt user to set template root
        """

        click.secho("Set template root", bg='blue', nl=True)

        templates_root_dir = click.prompt(
            text=click.style("Set path to template root dir: ", fg='bright_blue'),
            default=self.config['templates'].get('templates_root', None)
        )

        self.config['templates'].update({
            'templates_root': templates_root_dir
        })

    @staticmethod
    def _format_nodes(config):

        output = []

        for key_ in sorted(config):
            value_ = config[key_]
            if isinstance(value_, dict):
                value_ = yaml.dump(value_, indent=4, sort_keys=True)
            output.append({
                "name": key_, "message": value_
            })

        return output

    def get_state(self):
        """pretty prints a json config to terminal
        """
        config = self.config.copy()

        states = self._format_nodes(config)

        return states

    @property
    def local(self) -> dict:
        """get config object

        Returns:
            dict: config
        """
        return self.config.get('local', {})

    @property
    def workspace(self) -> dict:
        """get config object

        Returns:
            dict: config
        """
        return self.config.get('workspace', {})

    @property
    def templates(self) -> dict:
        """get config object

        Returns:
            dict: config
        """
        return self.config.get('templates', {})

class PackageConfigManager(BaseConfigManager):
    """Package Requirement Config Manager for sagify
    """

    def __init__(
        self,
        filepath: str,
        create: bool = False
    ):
        super().__init__()
        self.filepath = filepath

        self.config = self.load_config(self.filepath, create=create)

    @staticmethod
    def touch(path):
        """creat an empty txt file

        Args:
            filename (str): path to file
        """
        Path(path).touch()

    def load_config(self, file_name: str, create: bool) -> list:
        """load config

        Args:
            filename (str): path to config file to load

        Returns:
            list: config
        """
        self.check_if_exists_else_create(
            file_name=file_name, create=create
        )
        path = Path(file_name)
        with path.open() as file_:
            config = file_.readlines()

        for index, c_package in enumerate(config):
            if c_package.endswith("\n"):
                config[index] = config[index].split("\n")[0]
        return config

    def get_config(self) -> list:
        """get config object

        Returns:
            list: config
        """
        return self.config

    def seed_packages(self, packages: list):
        """seeds config with required package modules

        Args:
            packages (list): packages to add to config
        """
        for new_package in packages:
            for config_package in self.config:
                if new_package not in config_package:
                    self.config.append(new_package)

    def pretty_print(self):
        """pretty prints a list config to terminal
        """
        pretty_config = json.dumps(self.config, indent=4, separators=(',', ': '))
        logger.debug("{}\n".format(pretty_config))

    def write_file(self):
        """
        write to file
        """
        config = set(self.config)
        config_txt = "\n".join(config) + "\n"
        Path(self.filepath).write_text(config_txt)

class StageConfigManager(BaseConfigManager):
    """Development Stage Config Manager for mldock
    """

    def __init__(self, config: dict):
        self.config = config

    def ask_for_stages(self):
        """prompt user for stages
        """
        click.secho("Stages", bg='blue', nl=True)
        while True:
            echo_msg = click.style(
                "Add a development stage. ",
                fg='bright_blue'
            ) + "(Follow the prompts)"

            click.echo(
                echo_msg,
                nl=True
            )

            stage_name = click.prompt(
                text="Stage name: ",
                default="end",
                show_default=False,
                type=str
            )
            if stage_name == "end":
                break

            docker_tag = click.prompt(
                text="Set docker image tag: ",
                default=self.config.get(stage_name, {}).get('tag', None)
            )
            if stage_name not in self.config:
                self.config[stage_name] = {}

            self.config[stage_name].update(
                {'tag': docker_tag}
            )

class HyperparameterConfigManager(BaseConfigManager):
    """Hyperparameter Config Manager for mldock
    """

    def __init__(self, config: dict):
        self.config = config

    @staticmethod
    def is_float(value):
        """ Returns True is string is a number. """
        try:
            float(value)
            return True
        except ValueError:
            return False

    def ask_for_hyperparameters(self):
        """prompt user for hyperparameters
        """
        click.secho("Hyperparameters", bg='blue', nl=True)
        while True:

            echo_msg = click.style(
                "Add a hyperparameter. ",
                fg='bright_blue'
            )+"(Follow prompts)"

            click.echo(
                echo_msg,
                nl=True
            )
            hparam_name = click.prompt(
                text="Hyperparameter (name): ",
                default="end",
                show_default=False,
                type=str
            )

            if hparam_name == "end":
                break

            hparam_value = click.prompt(
                text="Set value: ",
                default=self.config.get(hparam_name, None)
            )

            self.config.update(
                {hparam_name: hparam_value}
            )

class EnvironmentConfigManager(BaseConfigManager):
    """Environment Config Manager for mldock
    """
    def __init__(self, config: dict):
        self.config = config

    @staticmethod
    def is_float(value):
        """ Returns True is string is a number. """
        try:
            float(value)
            return True
        except ValueError:
            return False

    @staticmethod
    def parse_bool(value):
        """ Returns True is string is a bool. """
        result = value
        if value.lower() == 'true':
            result = True
        elif value.lower() == 'false':
            result = False
        return result

    def ask_for_env_vars(self):
        """prompt user for environment variables
        """
        click.secho("Environment Variables", bg='blue', nl=True)
        while True:
            echo_msg = click.style(
                "Add a environment variable. ",
                fg='bright_blue'
            ) + "(Follow prompts)"

            click.echo(
                echo_msg,
                nl=True
            )

            env_var_name = click.prompt(
                text="Environment Variable (name): ",
                default="end",
                show_default=False,
                type=str
            )
            if env_var_name == "end":
                break

            env_var_value = click.prompt(
                text="Set value: ",
                default=self.config.get(env_var_name, None)
            )

            self.config.update(
                {env_var_name: env_var_value}
            )

class InputDataConfigManager(BaseConfigManager):
    """InputData Config Manager for mldock
    """
    file_ignores = []
    def __init__(self, config: dict, base_path: str):
        self.config = config
        self.base_path = base_path

    def write_gitignore(self):
        """
        write to file
        """
        config = {
            Path(dataset['channel'], dataset['filename']).as_posix()
            for dataset in self.config
        }

        config_txt = "\n".join(config) + "\n"

        ignore_filepath = Path(self.base_path,".gitignore")
        Path(ignore_filepath).write_text(config_txt)

    def ask_for_input_data_channels(self):
        """prompt user for hyperparameters
        """
        click.secho("Input Data Channels", bg='blue', nl=True)
        while True:

            echo_msg = click.style(
                "Add a data channel. ",
                fg='bright_blue'
            ) + "(Expects channel/filename). Hit enter to continue."

            channel_filename_pair = click.prompt(
                text=echo_msg,
                default="end",
                show_default=False,
                type=str
            )

            # pylint: disable=no-else-break
            if channel_filename_pair == "end":

                logger.debug("\nUpdated data channels")

                self.pretty_print()

                break

            elif "/" in channel_filename_pair and channel_filename_pair != 'channel/filename':

                channel, filename = channel_filename_pair.split("/", 1)

                logger.debug((
                    "Adding data/{CHANNEL}/{FILE_NAME}".format(
                        CHANNEL=channel,
                        FILE_NAME=filename
                    )
                ))

                self.config.append({
                    'channel': channel,
                    'filename': filename
                })

            else:
                logger.warning("Expected format as channel/filename. Skipping")

class ModelConfigManager(BaseConfigManager):
    """Model Config Manager for mldock
    """
    file_ignores = []
    def __init__(self, config: dict, base_path: str):
        self.config = config
        self.base_path = base_path

    def write_gitignore(self):
        """
        write to file
        """
        config = {
            Path(dataset['channel'], dataset['filename']).as_posix()
            for dataset in self.config
        }

        config_txt = "\n".join(config) + "\n"

        ignore_filepath = Path(self.base_path,".gitignore")
        Path(ignore_filepath).write_text(config_txt)

    def ask_for_model_channels(self):
        """prompt user for hyperparameters
        """
        click.secho("Model Channels", bg='blue', nl=True)
        while True:

            echo_msg = click.style(
                "Add a model channel. ",
                fg='bright_blue'
            ) + "(Expects channel/filename). Hit enter to continue."

            channel_filename_pair = click.prompt(
                text=echo_msg,
                default="end",
                show_default=False,
                type=str
            )

            # pylint: disable=no-else-break
            if channel_filename_pair == "end":
                logger.debug("\nUpdated model channels")
                self.pretty_print()

                break

            elif "/" in channel_filename_pair and channel_filename_pair != "channel/filename":
                channel, filename = channel_filename_pair.split("/", 1)
                logger.debug("Adding data/{}/{}".format(channel, filename))
                self.config.append({
                    'channel': channel,
                    'filename': filename
                })

            else:
                logger.warning("Expected format as channel/filename. Skipping")



"""Package utilities for managing dependencies with wheel"""

from __future__ import print_function
from __future__ import unicode_literals

import subprocess


def spawn(args, capture_output=False):
    """Spawn a subprocess to execute command"""
    print('=>', ' '.join(args))
    if capture_output:
        return subprocess.check_output(args)
    return subprocess.check_call(args)

def build_wheels(dist_dir, pip_wheel_args):
    """build wheels using pip wheel command line tool"""
    args = [
        'pip', 'wheel',
        '--wheel-dir', dist_dir
    ] + pip_wheel_args
    spawn(args)
    return dist_dir



"""LOGS API UTILITIES"""
import logging
from pathlib import Path
from clickclick.console import print_table
from pygrok import Grok
from pyarrow import fs
import gcsfs
import s3fs

from mldock.platform_helpers import utils

logger = logging.getLogger('mldock')

STYLES = {
    'FINE': {'fg': 'green'},
    'ERROR': {'fg': 'red'},
    'WARNING': {'fg': 'yellow', 'bold': True},
}


TITLES = {
    'state': 'Status',
    'creation_time': 'Creation Date',
    'id': 'Identifier',
    'desc': 'Description',
    'name': 'Name',
}

MAX_COLUMN_WIDTHS = {
    'desc': 50,
    'name': 20,
}

def infer_filesystem_type(path: str):
    """
        infers appropriate filesystem type and returns pyarrow support file-system

        args:
            path (str): full path/uri
        returns:
            file_system: a supported file-system type
            path: path without scheme
    """

    if utils._check_if_cloud_scheme(path, scheme=''):
        file_system = fs.LocalFileSystem()
    elif utils._check_if_cloud_scheme(path, scheme='s3'):
        path_without_scheme = utils.strip_scheme(path)
        file_system, path = s3fs.S3FileSystem(), path_without_scheme
    elif utils._check_if_cloud_scheme(path, scheme='gs'):
        path_without_scheme = utils.strip_scheme(path)
        file_system, path = gcsfs.GCSFileSystem(), path_without_scheme
    else:
        raise TypeError(
            "path scheme = '{SCHEME}' for '{PATH}' "
            "is not currently supported. "
            "Available options = 's3' or 'gs' "
            "or local filesystem path".format(
                SCHEME=utils.get_scheme(path),
                PATH=path
            )
        )
    return file_system, path

def read_file_stream(file_path, file_system):
    """
        Read file from file_system as a stream.

        args:
            file_path (str): path to file
            file_system (pyarrow.fs.FileSystem): pyarrow supported filesystem object
        return
            log_data (str): log data
    """
    if isinstance(file_system, fs.LocalFileSystem):
        with file_system.open_input_stream(file_path) as file_:
            log_data = file_.read().decode()
    else:
        with file_system.open(file_path, 'rb') as file_:
            log_data = file_.read().decode()
    return log_data

def parse_grok(file_path, pattern, file_system):
    """
        Parse logs using a custom crok pattern and return parsed objects.

        args:
            file_path (str): file path to object in file_system
            pattern (str): supported grok pattern
            file_system (pyarrow.fs.FileSystem): pyarrow supported file system object
        return:
            metadata (list): list of file objects as per grok pattern
    """
    grok = Grok(pattern)

    log_data = read_file_stream(file_path, file_system)

    metadata = grok.match(log_data.replace('\n', '\\n'))

    return metadata

def parse_grok_multiline(file_path, pattern, file_system):
    """
        Parse logs using a custom crok pattern and return parsed objects.

        args:
            file_path (str): file path to object in file_system
            pattern (str): supported grok pattern
            file_system (pyarrow.fs.FileSystem): pyarrow supported file system object
        return:
            metadata (list): list of file objects as per grok pattern
    """
    grok = Grok(pattern)

    metadata = []

    log_data = read_file_stream(file_path, file_system)

    for log in log_data.split('\n'):

        result = grok.match(log)
        if result is not None:
            metadata.append(result)

    return metadata

def get_all_file_objects(base_path, file_name, file_system):
    """
        Get all file objects, filter and return all files matching file_name

        args:
            base_path (str): base path to start walk from to find file objects
            file_name (str): file names to match and return
            file_system (pyarrow.fs.FileSystem): pyarrow supported file system object
        return:
            file_paths (list): list of file paths
    """

    if isinstance(file_system, fs.LocalFileSystem):
        file_selector = fs.FileSelector(base_path, recursive=True)
        logs = [
            log.path
            for log in file_system.get_file_info(file_selector)
            if log.base_name == file_name
        ]
    else:
        logs = [
            log
            for log in file_system.glob(f'{base_path}/**')
            if Path(log).name == file_name
        ]

    return logs

def extract_records_with_pattern(pattern, log_path, log_file):

    file_system, log_path = infer_filesystem_type(log_path)

    logs = get_all_file_objects(log_path, log_file, file_system)

    keys = set()
    rows = []
    for log in logs:
        run_id = Path(log).parents[0].name
        experiment = Path(log).parents[1].relative_to(log_path).as_posix().replace("/", ":")
        keys.add('run_id')
        keys.add('experiment')
        metadata = parse_grok_multiline(log, pattern, file_system)

        row = {}
        row.update({'run_id': run_id})
        row.update({'experiment': experiment})
        for obj in metadata:
            keys.add(obj['name'])
            row.update(
                {obj['name']: obj['value']}
            )
        rows.append(row)

    logger.info(metadata)
    print_table(keys, rows,
            styles=STYLES, titles=TITLES, max_column_widths=MAX_COLUMN_WIDTHS)



"""Container API methods"""

import os
from pathlib import Path
import logging
from github import Github
from environs import Env
from mldock.platform_helpers import utils

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('mldock')
env = Env()
env.read_env()  # read .env file, if it exists

def check_available_templates(
    templates_root,
    **kwargs
):
    """
        initialize mldock container project from template

        args:
            templates_root (str):
        kwargs:
            template_server (str): (default=local) template server to use
    """
    logger.info("Initializing Container Project")

    if kwargs.get('template_server', 'local') == 'local':
        template_dir = templates_root
        available_templates = [p.name for p in Path(template_dir).iterdir()]
    elif kwargs.get('template_server', 'local') == 'github':

        # set up in either ENV or configure local
        github_token = env("MLDOCK_GITHUB_TOKEN", default=None)
        github_org = env("MLDOCK_GITHUB_ORG", default='mldock')
        github_repo = env("MLDOCK_GITHUB_REPO", default='mldock-templates')
        github_branch = env("MLDOCK_GITHUB_REPO_BRANCH", default='main')

        github = Github(github_token)

        available_templates = utils.list_templates(
            github,
            '.',
            github_org,
            github_repo,
            github_branch,
            root=templates_root
        )
    
    return available_templates

def init_from_template(
    template_name,
    templates_root,
    src_directory,
    **kwargs
):
    """
        initialize mldock container project from template

        args:
            template_name (str):
            templates_root (str):
            src_directory (str):
            container_only (bool): (default=False) only initialize the container directory
            template_server (str): (default=local) template server to use
        kwargs:
            container_only (bool): (default=False) only initialize the container directory
            template_server (str): (default=local) template server to use
            prediction_script (str): (default=None) optional prediction script to use
            trainer_script (str): (default=None) optional trainer script to use
    """
    logger.info("Initializing Container Project")

    if kwargs.get('template_server', 'local') == 'local':
        template_dir = templates_root

    elif kwargs.get('template_server', 'local') == 'github':

        # set up in either ENV or configure local
        github_token = env("MLDOCK_GITHUB_TOKEN", default=None)
        github_org = env("MLDOCK_GITHUB_ORG", default='mldock')
        github_repo = env("MLDOCK_GITHUB_REPO", default='mldock-templates')
        github_branch = env("MLDOCK_GITHUB_REPO_BRANCH", default='main')

        github = Github(github_token)

        template_dir = utils.download_from_git(
            github,
            template_name,
            github_org,
            github_repo,
            github_branch,
            root=templates_root
        )

    logger.debug("Initializing template based on = {}".format(template_name))
    logger.debug("Template directory = {}".format(template_dir))

    logger.info("Setting up workspace")

    if kwargs.get('container_only', False):
        src_container_directory = os.path.join(
            src_directory,
            'container'
        )
        utils._copy_boilerplate_to_dst(
            os.path.join(template_dir, template_name, 'src/container'),
            src_container_directory,
            remove_first=True
        )
    else:
        utils._copy_boilerplate_to_dst(
            os.path.join(template_dir, template_name, 'src/'),
            src_directory
        )

    if kwargs.get('prediction_script', None) is not None:
        new_script_location = Path(src_directory, 'prediction.py')
        logger.info(f"Copying {kwargs.get('prediction_script', None)} => {new_script_location}")
        utils.copy_file(
            kwargs.get('prediction_script', None),
            new_script_location
        )

    if kwargs.get('trainer_script', None) is not None:
        new_script_location = Path(src_directory, 'trainer.py')
        logger.info(f"Copying {kwargs.get('trainer_script', None)} => {new_script_location}")
        utils.copy_file(
            kwargs.get('trainer_script', None),
            new_script_location
        )



"""
    Container Setup Utilities

    Implements workflows based on the environment and GCP specific code to
    download and upload ML assets e.g. Input data, Models and output data
"""
from functools import wraps
import traceback
from pathlib import Path

class BaseTrainingContainer:
    """
        A set of tasks for setup and cleanup of container
    """
    def __init__(self, container_environment, container_logger):
        self.container_logger = container_logger
        self.container_environment = container_environment

    def startup(self):
        """start up tasks executed on a container task setup"""
        self.container_logger.info("Setting up Instance")
        if self.container_environment.environment_variables('MLDOCK_STAGE', default=None) == "prod":
            self.container_logger.info("Env == Prod")
            self.container_environment.setup_inputs()
            self.container_environment.setup_model_artifacts()
        self.container_logger.info("Setup Complete")

    def cleanup(self):
        """clean up tasks executed on container task complete"""
        self.container_logger.info("Running Cleanup Script")
        if self.container_environment.environment_variables('MLDOCK_STAGE', default=None) == "prod":
            self.container_logger.info("Env == Prod")
            self.container_environment.cleanup_model_artifacts()
            self.container_environment.cleanup_outputs()
        self.container_logger.info("Cleanup Complete")

    def wrap(self, function):
        """execution handler for training container worflow"""
        @wraps(function)
        def func_wrapper(*args, **kwargs):
            """
                Wrapper function to decorate function
            """
            try:
                self.startup()
                return function(
                    *args,
                    self.container_environment,
                    self.container_logger,
                    **kwargs
                )
            except Exception as exception:
                # Write out an error file. This will be returned as the failureReason in the
                # DescribeTrainingJob result.
                trc = traceback.format_exc()
                log_file_path = Path(self.container_environment.output_data_dir, 'failure')
                with open(log_file_path, 'w') as file_:
                    file_.write('Exception during training: ' + str(exception) + '\n' + trc)
                # Printing this causes the exception to be in the training job logs, as well.
                self.container_logger.error('Exception during training: ' + str(exception) + '\n' + trc)
                raise
            finally:
                self.cleanup()

        return func_wrapper

class BaseServingContainer:
    """
        A set of tasks for setup and cleanup of container instance and
        between worker starts & completes
    """
    def __init__(self, container_environment, container_logger):
        self.container_environment = container_environment
        self.container_logger = container_logger

    def startup_worker(self):
        """steps to execute when a new worker starts an execution"""
        self.container_logger.info("Worker starting")

    def startup(self):
        """start up tasks executed on a container task setup"""
        self.container_logger.info("Setting up instance")

        if self.container_environment.environment_variables('MLDOCK_STAGE', default=None) == "prod":
            self.container_logger.info("Env == Prod")
            self.container_environment.setup_model_artifacts()
        self.container_logger.info("Setup Complete")

    def cleanup(self):
        """steps to execute when the machine shuts down"""
        # (Optional) instantiate super cleanup to maintain bases functionality
        self.container_logger.info("cleaning instance")

    def cleanup_worker(self):
        """steps to execute when a new worker completes an execution"""
        self.container_logger.info("Worker cleaning")

    def wrap(self, function):
        """execution handler for serving container worflow"""
        @wraps(function)
        def func_wrapper(*args, **kwargs):
            """
                Wrapper function to decorate function
            """
            try:
                self.startup_worker()

                return function(
                    *args,
                    self.container_environment,
                    self.container_logger,
                    **kwargs
                )
            except Exception as exception:
                # Write out an error file. This will be returned as the failureReason in the
                # DescribeTrainingJob result.
                trc = traceback.format_exc()
                log_file_path = Path(self.container_environment.output_data_dir, 'failure')
                with open(log_file_path, 'w') as file_:
                    file_.write('Exception during training: ' + str(exception) + '\n' + trc)
                # Printing this causes the exception to be in the training job logs, as well.
                self.container_logger.error('Exception during training: ' + str(exception) + '\n' + trc)
                raise
            finally:
                self.cleanup_worker()

        return func_wrapper



"""
    CONTAINER ASSETS MANAGEMENT

    Provides class objects with methods to execute at certain parts of the container life-cycle, i.e. startup or cleanup, etc.
    In short we call them TrainingContainer and ServingContainer classes. Each provides everything your container needs, including
    environment, logger, etc.
    Since Serving Containers and Training Containers are unique in use-case and requirements, we recommend providing one for each workflow.
"""
from pathlib import Path
import pandas as pd
from sklearn import datasets
from mldock.platform_helpers.mldock.configuration.container import BaseServingContainer, BaseTrainingContainer

# Training Container Assets Management
class TrainingContainer(BaseTrainingContainer):
    """
        Implements the base training container,
        allow a user to override/add/extend any training container setup logic
    """

    def startup(self):
        """steps to execute when the machine starts up"""
        # (Optional) instantiate super startup to maintain bases functionality
        super(TrainingContainer, self).startup()
        # add dev stage specific asset management tasks
        if self.container_environment.environment_variables('MLDOCK_STAGE', default=None) == "dev":
            # create channel if it doesn't exist
            data_channel = Path(self.container_environment.input_data_dir, 'iris')
            data_channel.mkdir(parents=True, mkdir=True)

            # download data
            iris = datasets.load_iris(as_frame=True)
            X = iris.data
            y = iris.target

            # write data to iris input data channel
            pd.concat(
                [X, y],
                axis=1
            ).to_csv(
                Path(data_channel, 'train.csv'),
                index=False
            )

# Serving Container Assets Management
class ServingContainer(BaseServingContainer):
    """
        Implements the base serving container,
        allow a user to override/add/extend any training container setup logic.
    """
    pass



"""PLATFORM HELPER UTILITIES"""
import sys
import os
import contextlib
import json
from urllib.parse import urlparse
from pathlib import Path
import tarfile
import zipfile
from typing import Iterator
import logging
import re
from distutils.dir_util import copy_tree, remove_tree
from distutils.dir_util import mkpath
from distutils.file_util import write_file, copy_file
import tempfile
import base64

from github import GithubException
from github.GithubException import UnknownObjectException

logger = logging.getLogger('mldock')

def _mkdir(dir_path: str):
    """make directory structure

    Args:
        dir_path (str): directory path to build
    """
    Path(dir_path).mkdir(parents=True, exist_ok=True)

def _iter_nested_dir(root_dir: str) -> Iterator[str]:
    """Iterate through nested folders.

    Args:
        root_dir (str): name of the directory to start form

    Yields:
        Iterator[str]: path iterator
    """
    rootdir = Path(root_dir)
    for path in rootdir.glob('**/*'):
        yield path

def _delete_file(file_path):
    """Delete a file

    Args:
        file_path (str): path to file

    Raises:
        TypeError: Expected a file_path of type file. Cannot be directory or other.
    """
    path = Path(file_path)
    if not path.is_file():
        raise TypeError("Expected a file_path of type file. Only deletes files")

    os.remove(path)

def _check_if_cloud_scheme(url: str, scheme: str = 's3') -> bool:
    parsed_url = urlparse(url)
    return parsed_url.scheme == scheme

def parse_url(url: str, scheme: str = 's3'):
    """Returns an (s3 bucket, key name/prefix) tuple from a url with an s3
    scheme.
    Args:
        url (str):
    Returns:
        tuple: A tuple containing:
            - str: S3 bucket name
            - str: S3 key
    """
    parsed_url = urlparse(url)
    if parsed_url.scheme != scheme:
        raise ValueError(
            "Expecting '{}' scheme, got: {} in {}.".format(
                scheme,
                parsed_url.scheme,
                url
            )
        )
    return parsed_url.netloc, parsed_url.path.lstrip("/")

def zip_folder(dir_path, output_file, rm_original=True):
    """zip in directory and optionally throw away unzipped"""

    with zipfile.ZipFile(output_file, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for file in _iter_nested_dir(dir_path):
            if file.name != Path(output_file).name and file.is_file():
                zipf.write(os.path.join(file), arcname=file.relative_to(dir_path))
                if rm_original:
                    _delete_file(file)

def zip_folder_as_tarfile(dir_path, output_file, rm_original=True):
    """zip folder as tarfile and optionally throw away original files"""
    with tarfile.open(output_file, "w:gz") as tar:
        for file in _iter_nested_dir(dir_path):
            if file.name != Path(output_file).name and file.is_file():
                tar.add(file, arcname=file.relative_to(dir_path))
                if rm_original:
                    _delete_file(file)

def unzip_file(filename, output_dir, rm_zipped=True):
    """unzip in directory and optionally throw away zipped"""
    with zipfile.ZipFile(filename, 'r', zipfile.ZIP_DEFLATED) as zipf:
        logger.info(
            "Unzipping {ZIP_FILE} => {OUTPUT_DIR}".format(
                ZIP_FILE=filename,
                OUTPUT_DIR=output_dir
            )
        )
        zipf.extractall(output_dir)
        if rm_zipped:
            logger.info("Removing {}".format(filename))
            _delete_file(filename)

def unzip_file_from_tarfile(filename, output_dir, rm_zipped=True):
    """untar in directory and optionally throw away zipped"""
    with tarfile.open(filename, "r:gz") as tar:
        logger.info("Unzipping {ZIP_FILE} => {OUTPUT_DIR}".format(
                ZIP_FILE=filename,
                OUTPUT_DIR=output_dir
            )
        )
        tar.extractall(output_dir)
        if rm_zipped:
            logger.info("Removing {FILE_PATH}".format(FILE_PATH=filename))
            _delete_file(filename)

def strip_scheme(url: str):
    """strip scheme from url"""
    parsed = urlparse(url)
    scheme = "{SCHEME}://".format(SCHEME=parsed.scheme)
    return parsed.geturl().replace(scheme, '', 1)

def get_scheme(path: str):
    """get the scheme of the given path"""
    parsed = urlparse(path)
    return parsed.scheme

def _read_json(file_path):
    """Read a JSON file.
    Args:
        path (str): Path to the file.
    Returns:
        (dict[object, object]): A dictionary representation of the JSON file.
    """
    with open(file_path, "r") as file_:
        return json.load(file_)

def _write_json(obj, file_path):
    """Write a serializeable object as a JSON file."""
    with open(file_path, "w") as file_:
        json.dump(obj, file_, indent=4)
        file_.write('\n')

def _write_file(file_path: str, parents: bool =True):
    """write a file"""
    if parents is True:
        mkpath(str(Path(file_path).parents[0].absolute()))
    write_file(str(Path(file_path).absolute()), '')

def _rename_file(base_path, current_filename, new_filename):
    """renames filename for a given base_path, saving the file in the same base_path

    Args:
        base_path ([type]): directory path containing file to rename
        current_filename ([type]): current name of the file to rename
        new_filename ([type]): new name for the renamed file
    """
    Path(base_path, current_filename).rename(Path(base_path, new_filename))

def _create_empty_file(base_path, filename):
    """renames filename for a given base_path, saving the file in the same base_path

    Args:
        base_path ([type]): directory path containing file to rename
        current_filename ([type]): current name of the file to rename
        new_filename ([type]): new name for the renamed file
    """
    Path(base_path, filename).touch(exist_ok=True)

def _copy_boilerplate_to_dst(src: str, dst: str, remove_first=False):
    """[summary]

    Args:
        src (str): [description]
        dst (str): [description]
    """
    source_path = Path(src).absolute().as_posix()
    destination_path = Path(dst).absolute().as_posix()
    if remove_first is True:
        try:
            logger.debug("removing first")
            remove_tree(destination_path)
        except FileNotFoundError:
            logger.debug("No file found. Assuming already deleted.")
    _ = copy_tree(source_path, destination_path)

def copy_file_to_new_file(
    infile,
    outfile,
    **kwargs
):
    """Copy a file respecting verbose, dry-run and force flags.  (The
    former two default to whatever is in the Distribution object, and
    the latter defaults to false for commands that don't define it.)"""
    return copy_file(
        infile,
        outfile,
        **kwargs
    )

@contextlib.contextmanager
def set_env(**environ):
    """
    Temporarily set the process environment variables.

    >>> with set_env(PLUGINS_DIR=u'test/plugins'):
    ...   "PLUGINS_DIR" in os.environ
    True

    >>> "PLUGINS_DIR" in os.environ
    False

    :type environ: dict[str, unicode]
    :param environ: Environment variables to set
    """
    old_environ = dict(os.environ)
    os.environ.update(environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)

def get_env_vars(environment, regex, flat=True):
    """Get all environ vars matching contains='<PREFIX>'"""
    if flat:
        filtered_env_vars = []
    else:
        filtered_env_vars = {}

    regex_pattern = re.compile(regex)

    for key,value in environment.items():
        result = regex_pattern.match(key)
        if result:
            # get all keys matching the regex
            if flat:
                filtered_env_vars.append({'key': key, 'value': value})
            else:
                filtered_env_vars.update({key: value})
    return filtered_env_vars

def get_sdk_credentials_volume_mount(
    auth_type='gcloud',
    **kwargs
):
    """
        configures a volume mount for providing provider resource
        permissions.
    """

    if auth_type == 'gcloud':
        host_path = kwargs.get('host_path', '~/.config/gcloud')
        container_path = kwargs.get('container_path', '/root/.config/gcloud')
    elif auth_type == 'awscli':
        host_path = kwargs.get('host_path', '~/.aws')
        container_path = kwargs.get('container_path', '/root/.aws')
    else:
        raise Exception("AUTH_TYPE = {} is not currently supported.")

    host_volume = Path(os.path.expanduser(host_path)).absolute().as_posix()
    container_bind = {'bind': container_path, 'mode': 'rw'}
    return {host_volume: container_bind}

def get_sha_for_tag(repository, tag):
    """
    Returns a commit PyGithub object for the specified repository and tag.
    """
    branches = repository.get_branches()
    matched_branches = [match for match in branches if match.name == tag]
    if matched_branches:
        return matched_branches[0].commit.sha

    tags = repository.get_tags()
    matched_tags = [match for match in tags if match.name == tag]
    if not matched_tags:
        raise ValueError('No Tag or Branch exists with that name')
    return matched_tags[0].commit.sha

def download_directory(repository, sha, server_path, local_prefix, relative_to):
    """
    Download all contents at server_path with commit tag sha in
    the repository.
    """
    try:
        contents = repository.get_contents(server_path, ref=sha)

        for content in contents:
            logger.info("Downloading {CONTENT_PATH}".format(CONTENT_PATH=content.path))
            local_path = Path(local_prefix, Path(content.path).relative_to(relative_to))
            if content.type == 'dir':
                local_path.mkdir(parents=True, exist_ok=True)
                download_directory(
                    repository=repository,
                    sha=sha,
                    server_path=content.path,
                    local_prefix=local_prefix,
                    relative_to=relative_to
                )
            else:
                file_content = repository.get_contents(content.path, ref=sha)
                file_data = base64.b64decode(file_content.content)

                # set executables with chmod +x
                # given that mldock containers always have the "executor.sh"
                if local_path.name == 'executor.sh':
                    local_path.touch(0o777)
                else:
                    local_path.touch()

                with open(local_path, "w+") as file_out:
                    file_out.write(file_data.decode())
                    file_out.write("\n")

    except (GithubException, IOError):

        # get exc info
        exc_type, _, _ = sys.exc_info()

        # if Unknown Object, inform user path is missing.
        if exc_type == UnknownObjectException:
            new_error = (
                "Server path = '{SERVER_PATH}' was "
                "not found in repository = '{REPOSITORY_NAME}'.".format(
                    SERVER_PATH=server_path,
                    REPOSITORY_NAME=repository.name
                )
            )
            logger.error(new_error)
            sys.exit(-1)
        raise

def download_from_git(github, start_dir, org, repo, branch, **kwargs):
    """Download the server path from git remote directory"""
    root = kwargs.get('root', '.')
    organization = github.get_organization(org)
    repository = organization.get_repo(repo)
    sha = get_sha_for_tag(repository, branch)

    tmp_dir = tempfile.mkdtemp()
    ## download to tmp
    download_directory(
        repository,
        sha,
        server_path=Path(root, start_dir).as_posix(),
        local_prefix=tmp_dir,
        relative_to=root
    )

    return tmp_dir

def list_objects_at_server_path(repository, sha, server_path):
    """
    Download all contents at server_path with commit tag sha in
    the repository.
    """
    try:
        contents = repository.get_contents(server_path, ref=sha)

        templates = []
        for content in contents:
            templates.append(Path(content.path).name)
        
        return templates

    except (GithubException, IOError):

        # get exc info
        exc_type, _, _ = sys.exc_info()
        raise

def list_templates(github, start_dir, org, repo, branch, **kwargs):
    """Download the server path from git remote directory"""
    root = kwargs.get('root', '.')
    organization = github.get_organization(org)
    repository = organization.get_repo(repo)
    sha = get_sha_for_tag(repository, branch)

    ## download to tmp
    return list_objects_at_server_path(
        repository,
        sha,
        server_path=Path(root, start_dir).as_posix()
    )



"""LOGS PARAMS COMMANDS"""
import logging
import click
from mldock.api.logs import extract_records_with_pattern

click.disable_unicode_literals_warning = True
logger = logging.getLogger('mldock')
MLDOCK_CONFIG_NAME = 'mldock.json'


@click.group()
def params():
    """
    Commands to manage and interact with params logs.
    """

@click.command()
@click.option(
    '--log-path',
    type=str,
    help='a grok pattern'
)
@click.option(
    '--log-file',
    type=str,
    default='logs.txt',
    help='file name'
)
def show(log_path, log_file):
    """show params for all runs as a table"""

    pattern=r"param: %{GREEDYDATA:name}=%{NUMBER:value};"

    extract_records_with_pattern(pattern, log_path, log_file)

params.add_command(show)



"""TEMPLATES COMMANDS"""
import logging
from pathlib import Path
import click

from mldock.platform_helpers import utils

click.disable_unicode_literals_warning = True
logger = logging.getLogger('mldock')
MLDOCK_CONFIG_NAME = 'mldock.json'


@click.group()
def templates():
    """
    Commands to create, update and manage container projects and templates.
    """

@click.command()
@click.option(
    '--name',
    help='Container name, used to name directories,etc.',
    required=True,
    type=str
)
@click.option(
    '--project_directory',
    '--dir',
    '-d',
    help='mldock container project.',
    required=True,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=False,
        allow_dash=False,
        path_type=None
    )
)
@click.option(
    '--out',
    help='Destination of template should be stored once created.',
    required=True,
    type=str
)
def create(name, project_directory, out):
    """
    Command to create a mldock enabled container template
    """

    try:
        if not Path(project_directory, MLDOCK_CONFIG_NAME).exists():
            raise Exception((
                "Path '{}' was not an mldock project. "
                "Confirm this directory is correct, otherwise "
                "create one.".format(project_directory)
            ))

        mldock_src_path = Path(project_directory, 'src')

        destination_path = Path(out, name)
        destination_path.mkdir(parents=False, exist_ok=True)

        destination_src_path = Path(destination_path, 'src')

        utils._copy_boilerplate_to_dst(mldock_src_path, destination_src_path)
    except Exception as exception:
        logger.error(exception)
        raise

templates.add_command(create)



"""LOGS COMMANDS"""
import logging
from pathlib import Path
import click
from clickclick import choice

from mldock.command.logs.metrics import metrics as metrics_commands
from mldock.command.logs.params import params as params_commands
from mldock.command.logs.grok import grok
from mldock.command.logs.errors import errors
from mldock.api.logs import read_file_stream, get_all_file_objects, infer_filesystem_type


click.disable_unicode_literals_warning = True
logger = logging.getLogger('mldock')
MLDOCK_CONFIG_NAME = 'mldock.json'

def reset_terminal():
    """clears the terminal view frame"""
    click.clear()

@click.group()
def logs():
    """
    Commands to manage and interact with logs.
    """

@click.command()
@click.option(
    '--log-path',
    type=str,
    help='a grok pattern'
)
@click.option(
    '--log-file',
    type=str,
    default='logs.txt',
    help='file name'
)
def show(log_path, log_file):
    """show errors for all runs as a table"""


    file_system, log_path = infer_filesystem_type(log_path)

    logs_data = get_all_file_objects(log_path, log_file, file_system)

    log_runs = [Path(log).parents[0].name for log in logs_data]

    state = choice('Select a run', log_runs, default=None)

    for log in logs_data:
        if Path(log).parents[0].name == state:
            log_file_path = log
            break

    log_data = read_file_stream(log_file_path, file_system)

    reset_terminal()
    print(log_data)

def add_commands(cli_group: click.group):
    """
        add commands to cli group
        args:
            cli (click.group)
    """
    cli_group.add_command(show)
    cli_group.add_command(metrics_commands)
    cli_group.add_command(params_commands)
    cli_group.add_command(grok)
    cli_group.add_command(errors)

add_commands(logs)
